XEmacs ICON PRELOAD "xemacs.ico"
XEmacsFile ICON "file.ico"
XEmacsLisp ICON "lisp.ico"

#include <windows.h>
#include "run.h"

1 VERSIONINFO 
FILEVERSION 1, 1, 4, 0
PRODUCTVERSION 1, 1, 4, 0
FILEOS VOS_NT_WINDOWS32
FILETYPE VFT_APP
FILESUBTYPE 0
BEGIN
 BLOCK "StringFileInfo"
 BEGIN
  BLOCK "04090000"
  BEGIN
   VALUE "CompanyName", "Charles S. Wilson\000"
   VALUE "FileDescription", "Runs console-mode programs sans console.\000"
   VALUE "FileVersion", "1.1.4\000"
   VALUE "InternalName", "run\000"
   VALUE "LegalCopyright", "Copyright © 1998 Charles S. Wilson\000"
   VALUE "OriginalFilename", "run.exe\000"
   VALUE "ProductName","run\000"
   VALUE "ProductVersion","1.1.4\000"
   VALUE "GPL Copyleft", "Released under the GNU General Public License version 2 --------------------------------------- This program is free software; you can redistribute it and/or modify it under the termms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. --------------------------------------- This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details. --------------------------------------- You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\000"
  END
 END
END
